#! perl

use v5.20;
use Test2::V0;

use CXC::Data::Visitor 'visit';
use Ref::Util 'is_refref', 'is_arrayref';
use Scalar::Util 'refaddr';
use experimental 'signatures', 'postderef';

sub myhash {
    (
        a => 1,
        b => [ 2, 3, 4 ],
        c => {
            d => [ 5, 6, 7 ],
            e => 8,
        },
    );
}

sub lookup ( $container, $kydx ) {
    return is_arrayref( $container ) ? $container->[$kydx] : $container->{$kydx};
}

sub check_ancestors ( $element, $meta ) {

    my @nodes   = ( $meta->{ancestors}->@*, $element );
    my @indexes = $meta->{path}->@*;

    my $node = shift @nodes;
    while ( @nodes ) {
        my $next  = shift @nodes;
        my $index = shift @indexes;
        is( lookup( $node, $index ), $next, "lookup: $index" );
        $node = $next;
    }
}

subtest 'default' => sub {

    my %hash = myhash;

    my @traversed;

    # see if context works
    my %context = ( traversed => \@traversed );

    visit(
        \%hash,
        sub ( $kydx, $vref, $context, $meta ) {

            subtest $kydx => sub {
                if ( is_refref( $vref ) ) {
                    my $container = lookup( $meta->{container}, $kydx );
                    is( refaddr( $container ), refaddr( $vref->$* ), 'container' );
                    push $context->{traversed}->@*, $kydx;
                }
                else {
                    subtest 'leaf' => sub {
                        my $element = lookup( $meta->{container}, $kydx );
                        is( $element, $vref->$*, "kydx: $kydx => $element" );

                        subtest 'ancestors' => sub {
                            check_ancestors( $element, $meta );
                        };
                    };

                    push $context->{traversed}->@*, [ $kydx, $vref->$* ];
                }
            };

            return 1;
        },
        context => \%context,
    );

    is(
        \@traversed,
        array {
            item [ a => 1 ];
            item 'b';
            item [ 0, 2 ];
            item [ 1, 3 ];
            item [ 2, 4 ];
            item 'c';
            item 'd';
            item [ 0, 5 ];
            item [ 1, 6 ];
            item [ 2, 7 ];
            item [ e => 8 ];
        },
        'scalars',
    );
};

subtest 'abort' => sub {

    my %hash = myhash;

    my @kydx;

    my ( $completed, $context, $meta ) = visit(
        \%hash,
        sub ( $kydx, $vref, $context, $meta ) {
            push @kydx, $kydx;
            return 0 if $kydx eq 'b';
            return 1;
        },
    );

    is( \@kydx, [ 'a', 'b', ] );
};

subtest 'cycle' => sub {

    subtest 'two parents' => sub {
        my %hash = ( a => { b => [ 0, 1 ] }, );

        $hash{c} = $hash{a};

        ok(
            lives {
                visit( \%hash, sub { 1 } )
            },
        ) or note $@;
    };

    subtest 'die' => sub {

        my %hash = ( a => { b => [ 0, 1 ] }, );
        $hash{a}{b}[0] = $hash{a};

        like(
            dies {
                visit( \%hash, sub { 1; }, )
            },
            qr/ cycle /,
        );
    };

    subtest 'truncate' => sub {

        my %hash = ( a => { b => [ 0, 1 ] }, );
        $hash{a}{b}[0] = $hash{a};

        ok(
            lives {
                visit( \%hash, sub { 1; }, cycle => 'truncate' )
            },
        ) or note $@;
    };

    subtest 'continue' => sub {

        my %hash = ( a => { b => [ 0, 1 ] }, );
        $hash{a}{b}[0] = $hash{a};

        my $seen = 0;
        ok(
            lives {
                visit(
                    \%hash,
                    sub ( $kydx, $vref, $context, $meta ) {
                        $seen++  if $kydx eq 'b';
                        return 0 if $seen == 10;
                        1;
                    },
                    cycle => 'continue',
                )
            },
        ) or note $@;
    };

};


done_testing;
